@@ -1,7 +1,7 @@
 package App::shcompgen;
 
-our $DATE = '2016-10-26'; # DATE
-our $VERSION = '0.27'; # VERSION
+our $DATE = '2016-10-28'; # DATE
+our $VERSION = '0.28'; # VERSION
 
 use 5.010001;
 use strict;
@@ -32,12 +32,12 @@ my $_complete_prog = sub {
         # user might want to mention a program file (e.g. ./foo)
         return {
             words => Complete::File::complete_file(
-                word=>$word, ci=>1, filter=>'d|rxf'),
+                word=>$word, filter=>'d|rxf'),
             path_sep => '/',
         };
     } else {
         # or user might want to mention a program in PATH
-        Complete::Program::complete_program(word=>$word, ci=>1);
+        Complete::Program::complete_program(word=>$word);
     }
 };
 
@@ -47,6 +47,11 @@ our %shell_arg = (
         summary => 'Override guessing and select shell manually',
         schema => ['str*', {in=>\@supported_shells}],
         tags => ['common'],
+        cmdline_aliases => {
+            fish => {summary=>"Shortcut for --shell=fish", is_flag=>1, code=>sub { $_[0]{shell} = "fish" }},
+            zsh  => {summary=>"Shortcut for --shell=zsh" , is_flag=>1, code=>sub { $_[0]{shell} = "zsh"  }},
+            tcsh => {summary=>"Shortcut for --shell=tcsh", is_flag=>1, code=>sub { $_[0]{shell} = "tcsh" }},
+        },
     },
 );
 our %common_args = (
@@ -135,6 +140,21 @@ _
         schema  => ['str*'],
         tags => ['common'],
     },
+
+    per_option => {
+        summary => 'Create per-option completion script if possible',
+        description => <<'_',
+
+If set to true, then attempt to create completion script that register each
+option. This creates nicer completion in some shells, e.g. fish and zsh. For
+example, option description can be shown.
+
+This is possible for only some types of scripts, e.g. <pm:Perinci::CmdLine>-
+(that does not have subcommands) or <pm:Getopt::Long::Descriptive>-based ones.
+
+_
+        schema => 'bool',
+    },
 );
 
 sub _all_exec_in_PATH {
@@ -225,6 +245,7 @@ sub _gen_completion_script {
     my $detres = $args{detect_res};
     my $shell  = $args{shell};
     my $prog   = $detres->[3]{'func.completee'} // $args{prog};
+    my $progpath = $args{progpath};
     my $qprog  = String::ShellQuote::shell_quote($prog);
     my $comp   = $detres->[3]{'func.completer_command'};
     my $qcomp  = String::ShellQuote::shell_quote($comp);
@@ -235,14 +256,14 @@ sub _gen_completion_script {
     my $script;
     my @helper_scripts;
 
-    if (($detres->[3]{'func.completer_type'} // '') eq 'Getopt::Long') {
+    if (($detres->[3]{'func.completer_type'} // '') =~ /\AGetopt::Long(?:::Descriptive)?\z/) {
         require Data::Dmp;
         require Getopt::Long::Dump;
 
-        my $progpath = $detres->[3]{'func.program_path'};
         my $content;
         my $dump_res = Getopt::Long::Dump::dump_getopt_long_script(
             filename => $progpath,
+            skip_detect => 1,
         );
         if ($dump_res->[0] != 200) {
             $log->errorf("Can't dump Getopt::Long script '%s': %s", $progpath, $dump_res);
@@ -298,11 +319,43 @@ complete -F _|."$prog $qprog".q|
 
     } elsif ($shell eq 'zsh') {
 
-        if (defined $args) {
-            $script = "# TODO: args not yet supported\n";
-        } else {
+      GEN_ZSH:
+        {
             $header_at_bottom++;
-            $script = q|#compdef |.$prog.q|
+            if ($args{per_option}) {
+                if ($detres->[3]{'func.completer_type'} =~ /^Perinci::CmdLine/) {
+                    require Complete::Zsh::Gen::FromPerinciCmdLine;
+                    my $res = Complete::Zsh::Gen::FromPerinciCmdLine::gen_zsh_complete_from_perinci_cmdline_script(
+                        filename => $progpath,
+                        skip_detect => 1,
+                    );
+                    if ($res->[0] == 200) {
+                        $log->debugf("Using per-option completion script for '%s'", $prog);
+                        $script = $res->[2];
+                        last GEN_ZSH;
+                    } else {
+                        $log->debugf("Can't generate per-option completion script for '%s': %s, falling back", $prog, $res);
+                    }
+                } elsif ($detres->[3]{'func.completer_type'} =~ /^Getopt::Long::Descriptive/) {
+                    require Complete::Zsh::Gen::FromGetoptLongDescriptive;
+                    my $res = Complete::Zsh::Gen::FromGetoptLongDescriptive::gen_zsh_complete_from_getopt_long_descriptive_script(
+                        filename => $progpath,
+                        skip_detect => 1,
+                    );
+                    if ($res->[0] == 200) {
+                        $log->debugf("Using per-option completion script for '%s'", $prog);
+                        $script = $res->[2];
+                        last GEN_ZSH;
+                    } else {
+                        $log->debugf("Can't generate per-option completion script for '%s': %s, falling back", $prog, $res);
+                    }
+                }
+            }
+
+            if (defined $args) {
+                $script = "# TODO: args not yet supported\n";
+            } else {
+                $script = q|#compdef |.$prog.q|
 _|.$prog.q|() {
     si=$IFS
     compadd -- $(COMP_SHELL=zsh COMP_LINE=$BUFFER COMP_POINT=$CURSOR |.$qcomp.q|)
@@ -310,6 +363,7 @@ _|.$prog.q|() {
 }
 _|.$prog.q| "$@"
 |;
+            }
         }
 
     } elsif ($shell eq 'tcsh') {
@@ -324,10 +378,43 @@ _|.$prog.q| "$@"
 
     } elsif ($shell eq 'fish') {
 
-        if (defined $args) {
-            $script = "# TODO: args not yet supported\n";
-        } else {
-            $script = "complete -c $qprog -a '(begin; set -lx COMP_SHELL fish; set -lx COMP_LINE (commandline); set -lx COMP_POINT (commandline -C); $qcomp; end)'\n";
+      GEN_FISH:
+        {
+            if ($args{per_option}) {
+                if ($detres->[3]{'func.completer_type'} =~ /^Perinci::CmdLine/) {
+                    require Complete::Fish::Gen::FromPerinciCmdLine;
+                    my $res = Complete::Fish::Gen::FromPerinciCmdLine::gen_fish_complete_from_perinci_cmdline_script(
+                        filename => $progpath,
+                        skip_detect => 1,
+                    );
+                    if ($res->[0] == 200) {
+                        $log->debugf("Using per-option completion script for '%s'", $prog);
+                        $script = $res->[2];
+                        last GEN_FISH;
+                    } else {
+                        $log->debugf("Can't generate per-option completion script for '%s': %s, falling back", $prog, $res);
+                    }
+                } elsif ($detres->[3]{'func.completer_type'} =~ /^Getopt::Long::Descriptive/) {
+                    require Complete::Fish::Gen::FromGetoptLongDescriptive;
+                    my $res = Complete::Fish::Gen::FromGetoptLongDescriptive::gen_fish_complete_from_getopt_long_descriptive_script(
+                        filename => $progpath,
+                        skip_detect => 1,
+                    );
+                    if ($res->[0] == 200) {
+                        $log->debugf("Using per-option completion script for '%s'", $prog);
+                        $script = $res->[2];
+                        last GEN_FISH;
+                    } else {
+                        $log->debugf("Can't generate per-option completion script for '%s': %s, falling back", $prog, $res);
+                    }
+                }
+            }
+
+            if (defined $args) {
+                $script = "# TODO: args not yet supported\n";
+            } else {
+                $script = "complete -c $qprog -a '(begin; set -lx COMP_SHELL fish; set -lx COMP_LINE (commandline); set -lx COMP_POINT (commandline -C); $qcomp; end)'\n";
+            }
         }
 
     } else {
@@ -422,13 +509,30 @@ sub _detect_prog {
 
   DETECT:
     {
-        if ($content =~
-                /^\s*# FRAGMENT id=shcompgen-hint command=(.+?)(?:\s+command_args=(.+))?\s*$/m
-                && $content !~ /^\s*# FRAGMENT id=shcompgen-nohint\s*$/m) {
-            # program give hints in its source code that it can be completed using a
-            # certain command
-            my $cmd = $1;
-            my $args = $2;
+        # split per line to avoid pathological case of slowness when number of
+        # source lines reaches many thousands.
+        my @lines = split /^/, $content;
+
+        my ($has_hint_cmd, $cmd, $args);
+        for my $line (@lines) {
+            if ($line =~
+                    /^\s*# FRAGMENT id=shcompgen-hint command=(.+?)(?:\s+command_args=(.+))?\s*$/) {
+                $has_hint_cmd++;
+                $cmd = $1;
+                $args = $2;
+                last;
+            }
+        }
+        my $has_nohint;
+        for my $line (@lines) {
+            if ($line =~ /^\s*# FRAGMENT id=shcompgen-nohint\s*$/) {
+                $has_nohint++;
+                last;
+            }
+        }
+        if ($has_hint_cmd && !$has_nohint) {
+            # program give hints in its source code that it can be completed
+            # using a certain command
             if (defined($args) && $args =~ s/\A"//) {
                 $args =~ s/"\z//;
                 $args =~ s/\\(.)/$1/g;
@@ -440,11 +544,18 @@ sub _detect_prog {
                 %extrametas,
             }];
         }
-        if(!$args{_skip_completer_hint} &&
-               $content =~
-               /^\s*# FRAGMENT id=shcompgen-hint completer=1 for=(.+?)\s*$/m
-               && $content !~ /^\s*# FRAGMENT id=shcompgen-nohint\s*$/m) {
-            my $completee = $1;
+
+        my $has_hint_completer;
+        my $completee;
+        for my $line (@lines) {
+            if ($line =~
+                    /^\s*# FRAGMENT id=shcompgen-hint completer=1 for=(.+?)\s*$/) {
+                $has_hint_completer++;
+                $completee = $1;
+                last;
+            }
+        }
+        if ($has_hint_completer && !$has_nohint) {
             return [400, "completee specified in '$progpath' is not a valid ".
                         "program name: $completee"]
                 unless $completee =~ $re_progname;
@@ -456,46 +567,28 @@ sub _detect_prog {
             }];
         }
 
-        if ($is_perl_script &&
-                # regex split here because i found a pathological case of very
-                # long matching time againt 'rsybak' datapacked script (~ 4M)
-                $content =~ /^\s*((?:use|require)\s+(Getopt::Long::Complete))\b/m ||
-                $content =~ /^\s*((?:use|require)\s+(Getopt::Long::Subcommand))\b/m) {
-            return [200, "OK", 1, {
-                "func.completer_command"=> $prog,
-                "func.completer_type"=> $2,
-                "func.note"=>"perl use/require statement: $1",
-            }];
-        }
+        if ($is_perl_script) {
+            for my $line (@lines) {
+                if ($line =~ /^\s*\# PERICMD_INLINE_SCRIPT: /) {
+                    # pericmd-inline script cannot complete themselves, but they
+                    # usually come with a separate completer script
+                    return [200, "OK", 0, {
+                        "func.reason" => "Perinci::CmdLine::Inline script",
+                    }];
+                }
+            }
 
-        if ($is_perl_script &&
-                $content =~ /^\s*((?:use|require)\s+(Getopt::Long))\b/m) {
-            return [200, "OK", 1, {
-                "func.program_path" => $progpath,
-                "func.completer_command" => $prog,
-                "func.completer_type" => $2,
-                "func.note"=>"perl use/require statement: $1",
-            }];
-        }
-
-      DETECT_PERICMD:
-        {
-            last unless $is_perl_script;
-            require Perinci::CmdLine::Util;
-            my $det_res = Perinci::CmdLine::Util::detect_perinci_cmdline_script(
-                string => $content);
-            $log->tracef("Perinci::CmdLine detection result: %s", $det_res);
-            last unless $det_res->[2];
-
-            # pericmd-inline doesn't currently support self-completion
-            last if $det_res->[3]{'func.is_inline'};
-
-            return [200, "OK", 1, {
-                "func.completer_command"=> $prog,
-                "func.completer_type"=> "Perinci::CmdLine",
-                "func.note"=>"detected using Perinci::CmdLine::Util",
-                "func.pericmd_detect_result" => $det_res,
-            }];
+            for my $line (@lines) {
+                if ($line =~ /^\s*((?:use|require)\s+
+                                  (Getopt::Long(?:::Complete|::Subcommand|::Descriptive)?|
+                                      Perinci::CmdLine(?:::Any|::Lite|::Classic)))\b/x) {
+                    return [200, "OK", 1, {
+                        "func.completer_command"=> $prog, # later will be set
+                        "func.completer_type"=> $2,
+                        "func.note"=>"perl use/require statement: $1",
+                    }];
+                }
+            }
         }
     }
     [200, "OK", 0];
@@ -557,7 +650,7 @@ sub _generate_or_remove {
             }
 
             my ($script, @helper_scripts) = _gen_completion_script(
-                %args, prog => $prog, detect_res => $detres);
+                %args, prog => $prog, progpath => $progpath, detect_res => $detres);
             my $comppath = _completion_script_path(
                 %args, prog => $prog, detect_res => $detres);
 
@@ -988,7 +1081,7 @@ _
                 my $res = list($args{args});
                 return undef unless $res->[0] == 200;
                 Complete::Util::complete_array_elem(
-                    array=>$res->[2], word=>$word, ci=>1);
+                    array=>$res->[2], word=>$word);
             },
         },
     },
@@ -1014,7 +1107,7 @@ App::shcompgen - Generate shell completion scripts
 
 =head1 VERSION
 
-This document describes version 0.27 of App::shcompgen (from Perl distribution App-shcompgen), released on 2016-10-26.
+This document describes version 0.28 of App::shcompgen (from Perl distribution App-shcompgen), released on 2016-10-28.
 
 =head1 FUNCTIONS
 
@@ -1095,6 +1188,17 @@ Directory to put helper scripts.
 
 Directory to put helper scripts.
 
+=item * B<per_option> => I<bool>
+
+Create per-option completion script if possible.
+
+If set to true, then attempt to create completion script that register each
+option. This creates nicer completion in some shells, e.g. fish and zsh. For
+example, option description can be shown.
+
+This is possible for only some types of scripts, e.g. L<Perinci::CmdLine>-
+(that does not have subcommands) or L<Getopt::Long::Descriptive>-based ones.
+
 =item * B<prog> => I<array[str]>
 
 Program(s) to generate completion for.
@@ -1223,6 +1327,17 @@ Directory to put helper scripts.
 
 Directory to put helper scripts.
 
+=item * B<per_option> => I<bool>
+
+Create per-option completion script if possible.
+
+If set to true, then attempt to create completion script that register each
+option. This creates nicer completion in some shells, e.g. fish and zsh. For
+example, option description can be shown.
+
+This is possible for only some types of scripts, e.g. L<Perinci::CmdLine>-
+(that does not have subcommands) or L<Getopt::Long::Descriptive>-based ones.
+
 =item * B<shell> => I<str>
 
 Override guessing and select shell manually.
@@ -1305,6 +1420,17 @@ Directory to put helper scripts.
 
 Directory to put helper scripts.
 
+=item * B<per_option> => I<bool>
+
+Create per-option completion script if possible.
+
+If set to true, then attempt to create completion script that register each
+option. This creates nicer completion in some shells, e.g. fish and zsh. For
+example, option description can be shown.
+
+This is possible for only some types of scripts, e.g. L<Perinci::CmdLine>-
+(that does not have subcommands) or L<Getopt::Long::Descriptive>-based ones.
+
 =item * B<shell> => I<str>
 
 Override guessing and select shell manually.
@@ -1385,6 +1511,17 @@ Directory to put helper scripts.
 
 Directory to put helper scripts.
 
+=item * B<per_option> => I<bool>
+
+Create per-option completion script if possible.
+
+If set to true, then attempt to create completion script that register each
+option. This creates nicer completion in some shells, e.g. fish and zsh. For
+example, option description can be shown.
+
+This is possible for only some types of scripts, e.g. L<Perinci::CmdLine>-
+(that does not have subcommands) or L<Getopt::Long::Descriptive>-based ones.
+
 =item * B<prog> => I<array[str]>
 
 Program(s) to remove completion script of.
